//
// Lighthouse3D.com Geometry Stereo Sample
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

#include <math.h>
#include <fstream>
#include <map>
#include <string>
#include <vector>


// include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Use Very Simple Libs
#include "../VSL/vslibs.h"
//#include <VSLibs/vslibs.h>



// a handle for the framebuffer object
GLuint fbo;
GLuint texFBO[2];
GLuint rb[2];

std::ostream *f;
VSMathLib *vsml;
VSShaderLib shader, shaderF, shaderS;
VSFontLib vsfl2, vsfl;
VSResModelLib myModel, myModel2, myPlane;
unsigned int aSentence, profileSentence;

GLuint texUnit;


// Camera Position
float camX, camY, camZ;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.5f;
float r = 5.0f;

// Mouse Tracking Variables
int startX, startY, tracking = 0;


// Frame counting and FPS computation
long myTime,timebase = 0,frame = 0;
char s[32];

#define FBOSIZE 512

// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h) {

	float ratio;
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
}


// ------------------------------------------------------------
//
// Render stuff
//



void renderScene(void) {

	{
		PROFILE("Frame");
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, fbo);


		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		


		changeSize(FBOSIZE, FBOSIZE);
		vsml->loadIdentity(VSMathLib::MODEL);
		
		// set camera
		float *aux, vecR[3],vecL[3] = {-camX, -camY, -camZ}, vecU[3] = {0.0f,1.0f,0.0f};
		vsml->loadIdentity(VSMathLib::VIEW);
		
		vsml->crossProduct(vecL, vecU, vecR);
		vsml->normalize(vecR);
		vsml->lookAt(camX-0.1f*vecR[0], camY-0.1f*vecR[1], camZ-0.1f*vecR[2], 0.0f,0.0f,0.0f, 0.0f,1.0f,0.0f);
		aux = vsml->get(VSMathLib::VIEW);
		shader.setUniform("view1", aux);

		vsml->loadIdentity(VSMathLib::VIEW);
		vsml->lookAt(camX+0.1*vecR[0], camY+0.1*vecR[1], camZ+0.1*vecR[2], 0,0,0, 0,1,0);
		aux = vsml->get(VSMathLib::VIEW);
		shader.setUniform("view2", aux);

		// use shader for pixel lighthing
		glUseProgram(shader.getProgramIndex());
		myModel2.render();




		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);


		// clear default frame buffer
		glViewport(0, 0, 1024, 512);
		glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glUseProgram(shaderS.getProgramIndex());
		// DRAW PLANE WITH TEXTURE FROM FBO
		glViewport(0, 0, 512, 512);
		myPlane.render();

		// DRAW SECOND PLANE WITH TEXTURE FROM FBO
		glViewport(512, 0, 512, 512);
		myPlane.render();
		


		//------------------------------------------
		// Done. Restoring Viewport
		changeSize(1024,512);
		glUseProgram(shaderF.getProgramIndex());

		// FPS computation and display
		frame++;
		myTime=glutGet(GLUT_ELAPSED_TIME);
		if (myTime - timebase > 1000) {
				sprintf(s,"FPS:%4.2f",
					frame*1000.0/(myTime-timebase));
			timebase = myTime;
			frame = 0;
			vsfl.prepareSentence(aSentence,s);
			glutSetWindowTitle(s);
		}

		{
			PROFILE("Dump");
			//set the shader for rendering the sentence
		

			vsfl.renderSentence(10,10,aSentence);
			std::string s = VSProfileLib::DumpLevels();
			vsfl.prepareSentence(profileSentence, s);
			vsfl.renderSentence(10, 30, profileSentence);

		}
		 //swap buffers
		{
			PROFILE("Swap");
			glutSwapBuffers();
		}
	} // end PROFILE("Frame")
	{
		PROFILE("Collect GL Queries Time");
		VSProfileLib::CollectQueryResults();
	}
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key) {

		case 27:

			glutLeaveMainLoop();
			break;

		case 'c': printf("Camera Spherical Coordinates (%f, %f, %f)\n", alpha, beta, r);
			break;
		case 'm': glEnable(GL_MULTISAMPLE); break;
		case 'n': glDisable(GL_MULTISAMPLE); break;
		case 'k': VSProfileLib::Reset(); break;
		case 'p': std::string s = VSProfileLib::DumpLevels();
				printf("%s\n", s.c_str());
				break;

	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//


void processMouseButtons(int button, int state, int xx, int yy)
{
	// start tracking the mouse
	if (state == GLUT_DOWN)  {
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
	}

	//stop tracking the mouse
	else if (state == GLUT_UP) {
		if (tracking == 1) {
			alpha -= (xx - startX);
			beta += (yy - startY);
		}
		else if (tracking == 2) {
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}


// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{

	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// left mouse button: move camera
	if (tracking == 1) {


		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// right mouse button: zoom
	else if (tracking == 2) {

		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


void mouseWheel(int wheel, int direction, int x, int y) {

	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}



// --------------------------------------------------------
//
// Shader Stuff
//


GLuint setupShaders() {

	// Shader for fonts
	shaderF.init();
	shaderF.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/color.vert");
	shaderF.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/color.frag");

	// set semantics for the shader variables
	shaderF.setProgramOutput(0,"outputF");
	shaderF.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shaderF.setVertexAttribName(VSShaderLib::TEXTURE_COORD_ATTRIB, "texCoord");

	shaderF.prepareProgram();
	VSGLInfoLib::getUniformsInfo(shaderF.getProgramIndex());

	// add sampler uniforms
	shaderF.setUniform("texUnit", 0);

	printf("InfoLog for Font Shader\n%s\n\n", shaderF.getAllInfoLogs().c_str());


	// Shader for stereo display
	shaderS.init();
	shaderS.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/showStereo.vert");
	shaderS.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/showStereo.frag");

	// set semantics for the shader variables
	shaderS.setProgramOutput(0,"outputF");
	shaderS.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shaderS.setVertexAttribName(VSShaderLib::TEXTURE_COORD_ATTRIB, "texCoord");

	shaderS.prepareProgram();
	VSGLInfoLib::getUniformsInfo(shaderS.getProgramIndex());

	// add sampler uniforms
	shaderS.setUniform("texUnit", 0);

	printf("InfoLog for Stereo Display Shader\n%s\n\n", shaderS.getAllInfoLogs().c_str());



	// Shader for models
	shader.init();

	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pixeldirdifambspec.vert");
	shader.loadShader(VSShaderLib::GEOMETRY_SHADER, "shaders/stereo.geom");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/pixeldirdifambspec.frag");

	// set semantics for the shader variables
	shader.setProgramOutput(0,"colorOut");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shader.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB,"normal");
	shader.setVertexAttribName(VSShaderLib::TEXTURE_COORD_ATTRIB, "texCoord");


	shader.prepareProgram();
	VSGLInfoLib::getUniformsInfo(shader.getProgramIndex());

	// add sampler uniforms
	shader.setUniform("texUnit", 0);

	float f[4] = {1.0f, 1.0f, 1.0f, 0.0f};
	shader.setBlockUniform("Lights","l_dir", f);
	
	printf("InfoLog for Model Shader\n%s\n\n", shader.getAllInfoLogs().c_str());
	
	return(shader.isProgramValid()&& shaderF.isProgramValid() && shaderS.isProgramValid());
}


// ------------------------------------------------------------
//
//  Framebufffer Object's stuff
//

void printFramebufferLimits() {

	int res;
	glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS, &res);
	printf("Max Color Attachments: %d\n", res);

	glGetIntegerv(GL_MAX_FRAMEBUFFER_WIDTH, &res);
	printf("Max Framebuffer Width: %d\n", res);

	glGetIntegerv(GL_MAX_FRAMEBUFFER_HEIGHT, &res);
	printf("Max Framebuffer Height: %d\n", res);

	glGetIntegerv(GL_MAX_FRAMEBUFFER_SAMPLES, &res);
	printf("Max Framebuffer Samples: %d\n", res);

	glGetIntegerv(GL_MAX_FRAMEBUFFER_LAYERS, &res);
	printf("Max Framebuffer Layers: %d\n", res);

	glGetIntegerv(GL_MAX_FRAMEBUFFER_LAYERS, &res);
	printf("Max Framebuffer Layers: %d\n", res);

}


void printFramebufferInfo(GLenum target, GLuint fbo) {

	glBindFramebuffer(target,fbo);

	int res;

	GLint buffer;
	int i = 0;
	do {
		glGetIntegerv(GL_DRAW_BUFFER0+i, &buffer);

		if (buffer != GL_NONE) {

			printf("Shader Output Location %d - color attachment %d\n", i, buffer - GL_COLOR_ATTACHMENT0);
			
			glGetFramebufferAttachmentParameteriv(target, buffer, GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE, &res);
			printf("\tAttachment Type: %s\n", res==GL_TEXTURE?"Texture":"Render Buffer");
			glGetFramebufferAttachmentParameteriv(target, buffer, GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME, &res);
			printf("\tAttachment object name: %d\n",res);
		}
		++i;

	} while (buffer != GL_NONE);

	glBindFramebuffer(target, 0);
}


GLuint createRGBATexture(int w, int h) {

	GLuint tex;

	glGenTextures(1, &tex);

	glBindTexture(GL_TEXTURE_2D_ARRAY, tex);

	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexStorage3D(GL_TEXTURE_2D_ARRAY, 1, GL_RGBA8, w, h, 2);

	glBindTexture(GL_TEXTURE_2D_ARRAY, 0);

	return(tex);
}

GLuint createDepthTexture(int w, int h) {

	GLuint tex;

	glGenTextures(1, &tex);

	glBindTexture(GL_TEXTURE_2D_ARRAY, tex);

	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexStorage3D(GL_TEXTURE_2D_ARRAY, 1, GL_DEPTH_COMPONENT32F, w, h, 2);

	glBindTexture(GL_TEXTURE_2D_ARRAY, 0);

	return(tex);
}

GLuint prepareFBO(int w, int h, int colorCount) {

	GLuint fbo;
	// Generate one frame buffer
	glGenFramebuffers(1, &fbo);
	// bind it
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, fbo);
	// attach textures for colors
	for (int i = 0; i < colorCount; ++i) {

		texFBO[i] = createRGBATexture(w,h);

		glFramebufferTexture(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0+i, texFBO[i], 0);
	}

	// attach depth texture
	GLuint depthFBO;
	depthFBO = createDepthTexture(w,h);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, depthFBO, 0);

	// check if everything is OK
	GLenum e = glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER);
	switch (e) {
	
		case GL_FRAMEBUFFER_UNDEFINED:
			printf("FBO Undefined\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT :
			printf("FBO Incomplete Attachment\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT :
			printf("FBO Missing Attachment\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER :
			printf("FBO Incomplete Draw Buffer\n");
			break;
		case GL_FRAMEBUFFER_UNSUPPORTED :
			printf("FBO Unsupported\n");
			break;
		case GL_FRAMEBUFFER_COMPLETE:
			printf("FBO OK\n");
			break;
		default:
			printf("FBO Problem?\n");
	}

	if (e != GL_FRAMEBUFFER_COMPLETE)
		return (0);
	
	// Set Drawing buffers
	GLuint attachments[8] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2,
							GL_COLOR_ATTACHMENT3, GL_COLOR_ATTACHMENT4, GL_COLOR_ATTACHMENT5,
							GL_COLOR_ATTACHMENT6, GL_COLOR_ATTACHMENT7};
	glDrawBuffers(colorCount,  attachments);

	// unbind fbo
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);

	return fbo;
}

// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//



int init()
{

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	// load models
	if (myModel2.load("../../models/teapot.3ds") && 
		myModel.load("../../models/cow.obj") &&
		myPlane.load("../../models/plane.obj")) {

		//printf("Models loaded\n%s", myModel.getInfo().c_str());

		// some GL settings
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
		glEnable(GL_MULTISAMPLE);

		glClearColor(0.1f, 0.1f, 0.1f, 0.0f);

		fbo = prepareFBO(FBOSIZE, FBOSIZE, 1);
		// fbo = 0 means that there was a problem building the fbo
		if (fbo == 0) {
			printf("Error in FBO creation\n");
			return false;
		}

		printFramebufferInfo(GL_DRAW_FRAMEBUFFER, fbo);

		myModel2.setTexture(0, texFBO[0], GL_TEXTURE_2D_ARRAY);
		myPlane.setTexture(0, texFBO[0], GL_TEXTURE_2D_ARRAY);
		return true;
	}
	else {
		printf("%s\n",myModel.getErrors().c_str());
		return false;
	}

}


void initVSL() {

	//std::ostream *f;
	//f = new std::ofstream("log.txt");

	//VSDebugLib::init(NULL);
	VSResourceLib::setMaterialBlockName("Material");

//	 Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
	vsml->setUniformName(VSMathLib::PROJECTION, "m_proj");
	vsml->setUniformName(VSMathLib::MODEL, "m_model");

	// Init VSFL Fonts
	vsfl.load("../../fonts/couriernew10");
	vsfl.setFixedFont(true);
	aSentence = vsfl.genSentence();
	profileSentence = vsfl.genSentence();

}

// ------------------------------------------------------------
//
// Main function
//


int main(int argc, char **argv) {

//  GLUT initialization
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion (4, 3);
	glutInitContextProfile (GLUT_CORE_PROFILE );
	//glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(1024,512);
	glutCreateWindow("Lighthouse3D - FBO Demo");

//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc ( mouseWheel ) ;

//	return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();
	if (glewIsSupported("GL_VERSION_4_2"))
		printf("Ready for OpenGL 4.2\n");
	else {
		printf("OpenGL 4.2 not supported\n");
		exit(1);
	}

   printf ("Vendor: %s\n", glGetString (GL_VENDOR));
   printf ("Renderer: %s\n", glGetString (GL_RENDERER));
   printf ("Version: %s\n", glGetString (GL_VERSION));
   printf ("GLSL: %s\n", glGetString (GL_SHADING_LANGUAGE_VERSION));
   int param;
   glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
   if (param == GL_CONTEXT_CORE_PROFILE_BIT)
	   printf("Context Profile: Core\n");
   else
	   printf("Context Profile: Compatibility\n");
   glGetIntegerv(GL_CONTEXT_FLAGS, &param);

   printFramebufferLimits();

	if (!setupShaders())
		exit(1);

	if (!init()) {
		exit(1);
	}

	initVSL();

	//  GLUT main loop
	glutMainLoop();

	return(1);

}

