#version 420

layout (std140) uniform Material {
	vec4 diffuse;
	vec4 ambient;
	vec4 specular;
	vec4 emissive;
	float shininess;
	int texCount;
};

uniform sampler2D texUnit;

in Data {
	vec3 normal;
	vec3 eye;
	vec3 lightDir;
	vec2 texCoord;
} DataIn;

out vec4 colorOut;

void main() {

	vec4 spec = vec4(0.0);

	vec3 n = normalize(DataIn.normal);
	vec3 l = normalize(DataIn.lightDir);
	vec3 e = normalize(DataIn.eye);

	float intensity = max(dot(n,l), 0.0);

	if (intensity > 0.0) {
		vec3 h = normalize(l + e);	
		float intSpec = max(dot(h,n), 0.0);
		spec = specular * pow(intSpec, shininess);
	}

	if (texCount == 0)
		colorOut = max(intensity *  diffuse + spec, ambient);
	else
		colorOut = max(intensity *  diffuse , ambient)* texture(texUnit, DataIn.texCoord) + spec;
	//colorOut = vec4(vec3(l),1.0);
	//colorOut = vec4(1.0,0.0,0.0,1.0);
}